# $Id: DBM.pl,v 1.8 2001/04/06 17:22:37 rschertler Exp $
#
# infobot :: Kevin Lenzo  (c) 1997

use strict;

package Infobot::DBM;

=head1 NAME

DBM.pl - infobot's interface to on-disk databases

=head1 SYNOPSIS

    openDBMx 'mydb', fatal => 1;	# more switches listed below

    $val = get 'mydb', $key;		# get value

    set 'mydb', $key, $val;		# set value
    $prev_val = postInc 'mydb', $key;	# increment, return old value
    $prev_val = postDec 'mydb', $key;	# decrement, return old value

    # run code with a lock in place, then put the previous lock back
    $code_returned = withDBMlock 'mydb', LOCK_EX, sub {
    	...
    };
    lockDBM 'mydb', LOCK_SH;		# explicit lock/unlock

    @keys = getDBMKeys 'mydb';		# get all keys
    clear 'mydb', $key;			# delete key
    clearAll 'mydb';			# delete all keys
    insertFile 'mydb', $filename;	# load space-separated fields
    closeDBM 'mydb';			# close this db
    closeDBMAll;			# close all dbs
    syncDBM 'mydb';			# flush changes to disk

=head1 DESCRIPTION

These functions provide B<infobot>'s interface to on-disk databases.

=cut

use vars qw(%DBMS $Debug $Init_done $Old_warnings);

use Fcntl	qw(
    :flock
    O_CREAT
    O_RDONLY
    O_RDWR
);

use Symbol	qw(
    gensym
);

use Util	qw(
    badinvo
    export_to_main
    import_from_main
    process_args
    uncontrol
);

BEGIN {
    if (!$Init_done) {
	$Old_warnings = $^W;
	$^W = 1;
    }
}

my @Import;
my @Export;

BEGIN {
    @Import = qw(
    	$filesep
	%param
	status
    );

    @Export = qw(
	clear
	clearAll
	closeDBM
	closeDBMAll
	forget
	get
	getDBMKeys
	insertFile
	lockDBM
	openDBM
	openDBMx
	postDec
	postInc
	set
	showdb
	syncDBM
	whatdbs
	withDBMlock
    );

    export_to_main @Export;
    import_from_main @Import;
}

use subs grep /^\w/, @Export;
use vars grep /^\W/, @Import, @Export;

use subs qw(_open lock);

# %DBMS maps from the user's database name to an array of data about each
# db.  The referenced array is indexed by the following constant subs.

%DBMS = () unless $Init_done;

sub F_DBNAME	() { 0 }	# %DBMS key
sub F_HASH	() { 1 }	# reference to the tied hash
sub F_FILE	() { 2 }	# name of file opened
sub F_LOCKING	() { 3 }	# true if locking is enabled for this db
sub F_LOCK_FILE	() { 4 }	# filehandle used for locking
sub F_LOCK_FH	() { 5 }	# filehandle used for locking
sub F_LOCK_STAT	() { 6 }	# current LOCK_* status
sub F_MODULE	() { 7 }	# database module used
sub F_SYNC_SUB	() { 8 }	# cached sync() method
sub F_INITFILE	() { 9 }	# initial contents when creating
sub F_UPDATE_COUNT () { 10 }	# number of updates since last sync

$Debug = 0 unless $Init_done;

my %Flock_desc = (
    (LOCK_SH)		=> 'bsh',
    (LOCK_SH | LOCK_NB)	=> 'nsh',
    (LOCK_EX)		=> 'bex',
    (LOCK_EX | LOCK_NB)	=> 'nex',
    (LOCK_UN)		=> 'bun',
    (LOCK_UN | LOCK_NB)	=> 'nun',
);

=head1 CONFIGURATION SETTINGS

=over 4

=item DBMModule

Setting C<DBMModule> lets you explicitly specify the DBM backend which
you'd like to use.  Standard values for this are C<NDBM_File>, C<DB_File>,
C<GDBM_File>, C<SDBM_File>, and C<ODBM_File>, but anything which provides
a tied hash interface should work.  If you don't specify this the default
will generally be the first of these which is present on your system.

Eg:

    DBMModule DB_File

=item DBMExt

This is appened to the file names passed to DBM open.  This can be
useful for DBM modules which don't modify the file name passed to them,
such as DB_File and GDBM_File.  For example,

    DBMExt .db

will provide traditional naming for DB_File databases.

=item sharedDBMs

This provides support for sharing database files among multiple B<infobot>s
on the same machine by using locking.  Eg, if you said

    sharedDBMs is are plusplus

your infobot would use locking when accessing the main factoid databases
and the C<karma> database, but not, say, the C<seen> database.  You can
have multiple infobots accessing the same databases for which they all use
locking.  It's up to you to make sure that all the bots which access a
particular file use locking for it, and that they use the same DBMModule
and DBMExt, if you screw any of these up the rogue will end up corrupting
your database.

There are two special values:

    sharedDBMs /all
    sharedDBMs /all-but-ignore

These set up locking for everything, and for everything but the
C<ignore> database (which is used more than any other, so perhaps
it's a good candidate for such special treatment).

=item commitDBM

This setting controls how often changes to the database are flushed to
disk.  Normally this isn't done manually, so it will depend on how the
DBM module you're using behaves.  If you set C<commitDBM> to a number,
changes will be forced to disk every that many updates (so use 1 to
force a sync after every update).

=back

=cut

unless ($Init_done) {
    $param{DBMModule}	= 'AnyDBM_File';
    $param{DBMExt}	= '';
    $param{sharedDBMs}	= '';
    $param{commitDBM}	= 0;
    # This is the extention added to generate a lock file name.  I'm
    # not documenting it as I'm not sure what sort of support I want to
    # do yet.  It might be nice to allow the user to lock one of the
    # database's files directly, but that might lead to deadlock.
    $param{lockFileExt}	= '.lf';
}

=head1 INTERFACE FUNCTIONS

=over 4

=item opemDBMx I<dbname>, [I<arg> => I<val>]...

This function opens up a database.  The I<dbname> is the name you'll use
to refer to it with all the other functions.  The normal practice is to
supply only the I<dbname>, most of the other arguments have reasonable
(preferred, even) defaults.  openDBMx() returns true if the database was
opened successfully, false otherwise, unless you've set C<fatal> to a
true value.

Arguments are:

    fatal => $boolean

This boolean, which is off by default, tells openDBMx() to
die() rather than returning false if the database can't be
successfully opened.

    tag => $tag

This defaults to the I<dbname> you gave.  That's normally what
you want, it'd be unusual to specify the tag manually.  The tag
is what's actually used to look up the other values in %param.

    file => $filename

This allows you to override the name of database file (though
the user's C<DBMExt> is still appended).  Normally you wouldn't
specify this, and the value the user specifies in $param{$tag}
is used.

    initfile => $filename

When a database is created the code uses insertFile() to load a
file called F<$misc_dir/infobot-$tag.txt> into it.  You can
override the name of the file used by specifying it with this
argument.

    locking => $boolean

This boolean tells the code whether to use locking or not.
Normally you wouldn't specify it and the user's C<sharedDBMs>
setting would dictate that.

    module => $db_module

This allows you to override the user's C<DBMModule> for this database.

=cut

sub openDBMx {
    my ($dbname, @arg) = @_;
    my ($fatal, $file, $initfile, $tag, $locking, $module);

    my $fail = sub {
	my $s = join '', @_;
	status $s;
	die $s if $fatal;
	return 0;
    };

    process_args \@arg,
	    fatal	=> \$fatal,
	    file	=> \$file,
	    locking	=> \$locking,
	    initfile	=> \$initfile,
	    module	=> \$module,
	    tag		=> \$tag
	or return;

    $tag ||= $dbname;

    if (!defined $file) {
	my $base = $param{$tag};
	if (!defined $base) {
	    return $fail->("$tag not specified in config file"
			    . " and no default supplied");
	}
	$file = $param{basedir} . $filesep . $base;
    }
    $file .= $param{DBMExt};

    $initfile = $param{confdir} . $filesep . "infobot-$tag.txt"
	if !defined $initfile;
    $locking = $param{sharedDBMs} eq '/all'
	    	|| ($tag ne 'ignore'
		    && $param{sharedDBMs} eq '/all-but-ignore')
		|| grep { $_ eq $tag } split ' ', $param{sharedDBMs}
	if !defined $locking;
    $module = $param{DBMModule} if !defined $module;

    eval "require $module";
    if ($@) {
	chomp $@;
	die "Invalid DBMModule setting `$module' ($@)\n";
    }
    # Simplify later code which needs to know the actual module.
    if ($module eq 'AnyDBM_File') {
	$module = $AnyDBM_File::ISA[0];
    }

    if ($DBMS{$dbname}) {
	status "$file replaces $DBMS{$dbname}[F_FILE]"
	    unless $file eq $DBMS{$dbname}[F_FILE];
    }

    my $rdb = $DBMS{$dbname} ||= [];
    $rdb->[F_DBNAME]	= $dbname;
    $rdb->[F_FILE]	= $file;
    $rdb->[F_LOCKING]	= $locking;
    $rdb->[F_MODULE]	= $module;
    $rdb->[F_INITFILE]	= $initfile;

    _open $rdb, 0
    	or return $fail->($@);

    # Make sure to release the lock, if the _open gave me one.
    lock $rdb, LOCK_UN;

    return 1;
}

# Translate the open flags argument for a DBM tie() as necessary.  Grr.

sub dbm_open_flags {
    @_ == 2 or badinvo;
    my ($module, $flags) = @_;

    if ($module eq 'GDBM_File') {
    	if ($flags == O_RDONLY)		{ return GDBM_File::GDBM_READER() }
    	elsif ($flags == O_RDWR)	{ return GDBM_File::GDBM_WRITER() }
    	elsif ($flags == O_CREAT|O_RDWR){ return GDBM_File::GDBM_WRCREAT() }
	else { badinvo 0, "unhandled flags $flags for $module" }
    }

    return $flags;
}

# Perform the actual open on the given db record.  Return true if
# successful, else false and set $@ to an explanation.

sub _open {
    @_ == 2 or badinvo;
    my ($rdb, $lock_bits) = @_;
    my ($do_status, $created);

    # Use exclusive access by default so non-shared dbs get opened
    # read/write.
    $lock_bits ||= LOCK_EX;

    my $lock_nb	= $lock_bits & LOCK_NB;
    my $lock_want = $lock_bits - $lock_nb;
    $lock_want == LOCK_SH
	or $lock_want == LOCK_EX
	or badinvo "invalid lock bits $lock_bits";

    my $dbname	= $rdb->[F_DBNAME];
    my $file	= $rdb->[F_FILE];
    my $locking	= $rdb->[F_LOCKING];
    my $module	= $rdb->[F_MODULE];

    # Do locking with an external file before the tie() to avoid caching
    # problems.  This file doesn't contain any data.  It's kept open all
    # the time, the actual DBM is open only while I've got a lock.
    $do_status = 1;
    if ($locking) {
    	# This happens over and over again, so skip the status output.
    	$do_status = $Debug;
	if (!$rdb->[F_LOCK_FH]) {
    	    my $lock_file = $file . $param{'lockFileExt'};
    	    $rdb->[F_LOCK_FILE] = $lock_file;

	    my $fh = $rdb->[F_LOCK_FH] = gensym;

	    if (!open $fh, "+>>$lock_file") {
		delete $DBMS{$dbname};
		$@ = "can't open locking filehandle on $lock_file for $dbname";
		return 0;
	    }

	    # I'm truncating the file to ensure that you don't use your
	    # database as the lock file, because I want to be able to
	    # support that better in the future.
	    truncate $fh, 0;

	    status "opened locking filehandle for $dbname -> $lock_file";
	}
	if (!flock $rdb->[F_LOCK_FH], $lock_bits) {
	    $@ = "can't lock $rdb->[F_LOCK_FILE]: $!";
	    return 0;
	}
    }

    # Now I've got the lock, so I have to release it if I return failure.
    my $unlock = sub {
	if ($locking) {
	    flock $rdb->[F_LOCK_FH], LOCK_UN | $lock_nb
		or die "Can't unlock database $dbname: $!";
	}
    };

    my $opened = 0;
    if ($lock_want == LOCK_SH) {
	if (tie %{ $rdb->[F_HASH] }, $module, $file,
		dbm_open_flags($module, O_RDONLY), 0) {
    	    $opened = 1;
	    status "opened $dbname -> $file (read only)"
		if $do_status;
	}
	else {
	    # I used a shared lock, but to create I'll need an exclusive
	    # lock.
	    $lock_want = LOCK_EX;
	    $lock_bits = $lock_want | $lock_nb;
	    if ($locking && !flock $rdb->[F_LOCK_FH], $lock_bits) {
		$@ = "can't lock $rdb->[F_LOCK_FILE]: $!";
		$unlock->();
		return 0;
	    }
	}
    }
    else {
	if (tie %{ $rdb->[F_HASH] }, $module, $file,
		dbm_open_flags($module, O_RDWR), 0) {
	    $opened = 1;
	    status "opened $dbname -> $file"
		if $do_status;
	}
    }

    # Create the database if it wasn't there.
    if (!$opened) {
	if (tie %{ $rdb->[F_HASH] }, $module, $file,
	    	dbm_open_flags($module, O_CREAT | O_RDWR), 0666) {
	    status "created new db $dbname -> $file";
	    $created = 1;
	}
	else {
	    $@ = "failed to open $dbname -> $file ($!)";
	    $unlock->();
	    return 0;
	}
    }

    $rdb->[F_LOCK_STAT] = $lock_want;
    $rdb->[F_UPDATE_COUNT] = 0;
    insertFile $dbname, $rdb->[F_INITFILE]
	if $created;

    return 1;
}

sub _close_open {
    my ($dbname) = @_;
    my ($fail_reason);

    closeDBM '_no_delete', $dbname;

    # The old (commented-out) code for this would sleep and retry if the
    # reopen failed.  It seems bogus to me, but I don't want to piss
    # anybody off by removing it.

    for (1..10) {
	return 1 if _open $DBMS{$dbname}, 0;
    } continue {
	status "Error re-opening $dbname ($@), sleeping";
	sleep 1;
    }
    status "Error re-opening $dbname ($@), giving up";
    return 0;
}

=item openDBM $dbname => $file, ...

This is the old interface to opening databases.  It's equivalent to
running

    openDBMx $dbname, file => $file;

for each pair of arguments.  The return value is true if all the opens
succeeded.

=cut

sub openDBM {
    my %arg = @_;
    my ($dbname, $file, $fail);

    while (($dbname, $file) = each %arg) {
	next unless $dbname =~ /\S/;
	openDBMx $dbname, file => $file
	    or $fail = 1;
    }

    return !$fail;
}

=item syncDBM $dbname

Flush to disk any unwritten changes to the database.

=cut

sub syncDBM {
    my ($dbname) = @_;
    my $rdb = $DBMS{$dbname};

    print "sync $rdb->[F_DBNAME]\n" if $Debug;
    $rdb->[F_UPDATE_COUNT] = 0;

    # If locking is active the database might not actually be open (if
    # called from outside this file, that shouldn't happen when called
    # internally).
    if (!$rdb->[F_HASH]) {
	print "syncDBM: $dbname skipping sync\n" if $Debug;
	return;
    }

    &{ $rdb->[F_SYNC_SUB] ||= do {
		if (tied(%{ $rdb->[F_HASH] })->can('sync')) {
		    print "syncDBM: $dbname using ->sync\n" if $Debug;
		    sub { tied(%{ $rdb->[F_HASH] })->sync }
		}
		else {
		    print "syncDBM: $dbname using reopen\n" if $Debug;
		    sub { _close_open $dbname }
		}
	    }
	}();
}

sub lock {
    my ($rdb, $bits, $caller_level) = @_;

    my $have	= $rdb->[F_LOCK_STAT];
    my $nb	= $bits & LOCK_NB;
    my $want	= $bits - $nb;

    if ($Debug) {
	my $caller = join ':', (caller 1 + $caller_level)[1, 2];
	$caller =~ s-.*/--;
	printf "lock db %-8s fd %2s from %s -> %s bits %s caller %s\n",
		$rdb->[F_DBNAME],
		$rdb->[F_LOCK_FH] ? fileno $rdb->[F_LOCK_FH] : '-',
		$Flock_desc{$have} || $have,
		$Flock_desc{$want} || $want,
		$Flock_desc{$bits} || $bits,
		$caller;
    }

    return if $have == $want;

    # Possibly flush when unlocking (or downgrading LOCK_EX to LOCK_SH).
    if ($have == LOCK_EX) {
	$rdb->[F_UPDATE_COUNT]++;
	if ($param{commitDBM} eq 'ALWAYS' # grandfather
                || ($param{commitDBM} > 0 &&
                    $rdb->[F_UPDATE_COUNT] >= $param{commitDBM})) {
	    syncDBM $rdb->[F_DBNAME]
		unless $rdb->[F_LOCKING]; # no point, about to untie
	}
    }

    if (!$rdb->[F_LOCKING]) {
	$rdb->[F_LOCK_STAT] = $want;
	return;
    }

    # I can't directly upgrade or downgrade a lock, because I open the
    # database in different modes.  Simulate by re-opening.

    if ($have != LOCK_UN) {
	closeDBM '_no_delete', $rdb->[F_DBNAME];
	flock $rdb->[F_LOCK_FH], LOCK_UN | $nb
	    or die "Can't unlock $rdb->[F_FILE]: $!\n";
	$have = $rdb->[F_LOCK_STAT] = LOCK_UN;
    }

    if ($want == LOCK_UN) {
	return;
    }

    _open $rdb, $bits
	or die "Failed to open $rdb->[F_DBNAME] at lock time: $@";
}

=item lockDBM $dbname, $lock_mode

Lock the given database with the given flock()-style mode.

=cut

sub lockDBM {
    @_ == 2 or badinvo;
    my ($dbname, $bits) = @_;

    lock $DBMS{$dbname}, $bits;
}

sub with_lock {
    @_ == 3 or badinvo;
    my ($rdb, $bits, $code) = @_;

    my $nb = $bits & LOCK_NB;
    my $want = $bits - $nb;
    $want == LOCK_SH
	or $want == LOCK_EX
	or badinvo "invalid lock bits $bits";

    # Only do locking if upgrading (don't go from LOCK_EX to LOCK_SH,
    # and don't change an existing lock of the right type).
    my $do_lock = 0;
    my $had = $rdb->[F_LOCK_STAT];
    if ($had != LOCK_EX && $had != $want) {
    	$do_lock = 1;
	lock $rdb, $bits, 1;
    }

    my @ret;
    if (!defined wantarray) {
	$code->();
    }
    else {
	@ret = wantarray ? $code->() : (scalar $code->());
    }

    lock $rdb, $had | $nb, 1 if $do_lock;

    return wantarray ? @ret : $ret[0];
}

=item withDBMlock $dbname, $lock_mode, $code

Lock the given database with the given flock()-style mode, run $code,
then put the lock back to how it was.  The return value is whatever the
$code returns.

=cut

sub withDBMlock {
    @_ == 3 or badinvo;
    my ($dbname, $bits, $code) = @_;

    with_lock $DBMS{$dbname}, $bits, $code;
}

=item insertFile $dbname, $filename

This loads the given file into the database.  Input lines look like

    key => value

(spaces around the C<=E<gt>> are optional).

=cut

sub insertFile {
    my ($dbname, $factfile) = @_;
    my $rdb = $DBMS{$dbname};

    if (open(IN, $factfile)) {
	my ($good, $total);

	with_lock $rdb, LOCK_EX, sub {
	    while(<IN>) {
		chomp;
		my ($k, $v) = split(/\s*=>\s*/, $_, 2);
		if ($k and $v) {
		    $rdb->[F_HASH]{$k} = $v;
		    $good++;
		}
		$total++;
	    }
	};
	close(IN);
	status "loaded $factfile into $dbname ($good/$total good items)";
    } else {
	status "FAILED to load $factfile into $dbname";
    }
}

=item closeDBM $dbname

Close the database.

=cut

sub closeDBM {
   if (@_) {
	my ($dbname, $rdb, $no_delete);
	$no_delete = shift if $_[0] eq '_no_delete';
	foreach $dbname (@_) {
	    my $rdb = $DBMS{$dbname};
	    my $rhash = $rdb->[F_HASH];
	    my $do_status = 1;

    	    # If not deleting the %DBMS entry, at least delete the
    	    # F_HASH element so it's possible to tell if a %DBMS entry
    	    # is currently tied.
	    if ($no_delete) {
	    	$rdb->[F_HASH] = undef;
		# This happens all the time when locking, so don't
		# output about it.
		$do_status = $Debug if $rdb->[F_LOCKING];
	    }
	    else {
		delete $DBMS{$dbname};
	    }

	    my $ok = untie(%$rhash);
	    status $ok
		    ? "closed db $dbname"
		    : "Error closing db $dbname ($!)"
		if $do_status;
	}
    } else {
	status "No dbs specified; none closed";
    }
}

=item closeDBMAll

Close all databases.

=cut

sub closeDBMAll {
    closeDBM keys %DBMS;
}

=item set $dbname, $key, $val

Set a key/value pair in the database.

=cut

sub set {
    my ($dbname, $key, $val) = @_;

    if (!$key) {
	($dbname, $key, $val) = split(/\s+/, $dbname);
    }
    printf "set $dbname %s => %s\n", uncontrol($key), uncontrol($val)
	if $Debug;

    # this is a hack to keep set param consistant.. overloaded
    if ($dbname eq 'param') {
	my $was = $param{$key};
	$param{$key} = $val;
	return $was;
    }

    if (!$key) {
	return 'NULLKEY';
    }

    my $rdb = $DBMS{$dbname};
    return with_lock $rdb, LOCK_EX, sub {
	my $was = $rdb->[F_HASH]{$key};
	$rdb->[F_HASH]{$key} = $val;
	return $was;
    };
}

=item get $dbname, $key

Return the value corresponding to the $key in the database.

=cut

sub get {
    my ($dbname, $key) =@_;

    if (!$key) {
	($dbname, $key) = split(/\s+/, $dbname);
    }

    my $rdb = $DBMS{$dbname};
    return with_lock $rdb, LOCK_SH, sub {
	return $rdb->[F_HASH]{$key};
    };
}

=item postInc $dbname, $key

Increment the value of $key in the database, return the old value.

=cut

sub postInc {
    my ($dbname, $key) = @_;

    my $rdb = $DBMS{$dbname};
    return with_lock $rdb, LOCK_EX, sub {
	return set $dbname, $key, 1 + get $dbname, $key;
    };
}

=item postDec $dbname, $key

Decrement the value of $key in the database, return the old value.

=cut

sub postDec {
    my ($dbname, $key) = @_;

    my $rdb = $DBMS{$dbname};
    return with_lock $rdb, LOCK_EX, sub {
	return set $dbname, $key, -1 + get $dbname, $key;
    };
}

sub whatdbs {
    my @result;
    foreach (keys %DBMS) {
	push @result, "$_ => $DBMS{$_}[F_FILE]";
    }
    return @result;
}

sub showdb {
    my ($dbname, $regex) = @_;
    my @result;

    if (!$regex) {
	($dbname, $regex) = split(/\s+/, $dbname, 2);
    }

    my @whichdbs;

    if (!$dbname) {
	status "no db given";
	status "try showdb <db> <regex>";
	# @whichdbs = (keys %DBMS);
    } else {
	@whichdbs = ($dbname);
    }

    foreach $dbname (@whichdbs) {
	my $rdb = $DBMS{$dbname};
	if (!$rdb) {
	    status "the database $dbname is not open.";
	    status "try showdb <db> <regex>";
	    return();
	}
	with_lock $rdb, LOCK_SH, sub {
	    my $rhash = $rdb->[F_HASH];
	    my ($key, $val);
	    if (!$regex) {
		status "showing all of $dbname";
		while (($key, $val) = each %$rhash) {
		    push @result, "$key => $val";
		}
	    } else {
		status "searching $dbname for /$regex/";
		while (($key, $val) = each %$rhash) {
		    push @result, "$key => $val"
			if $key =~ /$regex/ || $val =~ /$regex/;
		}
	    }
	};
    }

    return @result;
}

sub forget {
    clear @_;
    return '';
}

=item clear $dbname, $key

Delete a key from the database.

=cut

sub clear {
    my ($dbname, $key) =@_;

    if (!$key) {
	($dbname, $key) = split(/\s+/, $dbname);
    }

    my $rdb = $DBMS{$dbname};
    return with_lock $rdb, LOCK_EX, sub {
	print "DELETING $dbname $key\n";
	my $was = delete $rdb->[F_HASH]{$key};
	print "DELETED\n";
	return $was;
    }
}

=item clearAll $dbname

Empty the database.

=cut

sub clearAll {
    my ($dbname) = @_;

    my $rdb = $DBMS{$dbname};
    with_lock $rdb, LOCK_EX, sub {
	%{ $rdb->[F_HASH] } = ();
    };
}

=item getDBMKeys $dbname

Return all the keys in the database.

=cut

sub getDBMKeys {
    my ($dbname) = @_;

    my $rdb = $DBMS{$dbname};
    return with_lock $rdb, LOCK_SH, sub {
	return keys %{ $rdb->[F_HASH] };
    };
}

if (!$Init_done) {
    $^W = $Old_warnings;
    $Init_done = 1;
}

1

__END__

=back

=head1 AUTHOR

Kevin Lenzo, expanded by Roderick Schertler <F<roderick@argon.org>>

=head1 SEE ALSO

infobot(1), perl(1).

=cut
