# $Id: Util.pm,v 1.2 2001/04/06 17:22:37 rschertler Exp $

use strict;

package Util;

=head1 NAME

Util - infobot utility functions

=head1 SYNOPSIS

    export_to_main   qw(func &func2 $scalar @array %hash);
    import_from_main qw(func &func2 $scalar @array %hash);
    import_export $from_pkg, $to_pkg, @symbol;

    process_args \@arg_list, myarg => \$myvar, %more_pairs or die;

=head1 DESCRIPTION

This module provides some utility functions for the B<infobot>.

=cut

use Carp		qw(croak);
use Exporter		();

use vars qw($VERSION @ISA @EXPORT_OK);

$VERSION  = do{my@r=q$Revision: 1.2 $=~/\d+/g;sprintf '%d.'.'%03d'x$#r,@r};

@ISA		= qw(Exporter);
@EXPORT_OK	= qw(
    badinvo
    export_to_main
    fileline
    import_export
    import_from_main
    process_args
    subcall_info
    subname
    uncontrol
);

use subs @EXPORT_OK;

=head1 IMPORTABLE SYMBOLS

=over 4

=cut

sub import_export {
    my ($from_pkg, $to_pkg, @symbol) = @_;
    my ($symbol, $type, $name, $code);

    $code = "package $to_pkg;\n";
    for $symbol (@symbol) {
	($type, $name) = $symbol =~ /^([\$\@%&])?(\w+)$/
	    or croak "Invalid symbol `$symbol'";
	$type ||= '&';
	$code .= "*$name = \\$type${from_pkg}::$name;\n";
    }
    print $code if 0;

    {
	no strict 'refs';
	eval $code;
	die if $@;
    }
}

sub export_to_main {
    my @symbol = @_;
    import_export scalar(caller), 'main', @symbol;
}

sub import_from_main {
    my @symbol = @_;
    import_export 'main', scalar(caller), @symbol;
}

#------------------------------------------------------------------------------

BEGIN {
    import_from_main qw(status);
}

=item B<badinvo> [I<level> [I<message>]]

Die with a message indicating that the current subroutine was invoked
improperly.  If I<level> is given go back that many additional levels
before the current one.  If I<message> is also given include it in the
die message.

=cut

sub badinvo {
    my $level = @_ ? shift : 0;
    my $msg = @_ ? " (" . shift() . ")" : '';
    die "Bogus invocation of ", subname($level + 1),
        " from ", subcall_info($level + 1), "$msg\n";
}

=item B<fileline> [I<level>]

Return a string describing the caller's file and line number.  If
I<level> is given it's an additional number of stack frames to go
back.

=cut

sub fileline {
    @_ <= 1 || croak 'bogus invocation of ', (caller 0)[3];
    my $level = shift || 0;
    my ($package, $file, $line) = caller $level;
    return "$file line $line";
}

sub process_args {
    my ($rarg, %desc) = @_;

    my $caller_sub = (caller 1)[3];
    my $fail = 0;

    while (@$rarg > 1) {
	my ($key, $val) = splice @$rarg, 0, 2;
	if ($desc{$key}) {
	    ${ $desc{$key} } = $val;
	} else {
	    status "$caller_sub: invalid arg `$key'";
	    $fail = 1;
	}
    }

    if (@$rarg) {
	status "$caller_sub: ignoring trailing value-less arg `$rarg->[0]'";
	$fail = 1;
    }

    return !$fail;
}

=item B<subname> [I<level>]

Return the name of the current subroutine (that is, the one which invokes
B<subname>).  If I<level> is given, go back that many additional stack
frames and give the name that sub instead.

=cut

sub subname {
    @_ <= 1 || croak 'bogus invocation of ', (caller 0)[3];
    my $level = shift || 0;
    return +(caller $level + 1)[3];
}

=item B<subcall_info> [I<level>]

Return a text string which describes the invocation of the current
subroutine.  Like B<subname>, current really means B<subcall_info>'s
caller, and you can specify a I<level> to go back to a different stack
frame.

=cut

#'

sub subcall_info {
    @_ <= 1 || croak 'bogus invocation of ', (caller 0)[3];
    my $level = shift || 0;
    my $fileline = fileline $level + 2;
    # The $subname in that stack frame is the name of the sub invoked,
    # not the invoker.  So, the invoker is one frame further up.
    my $sub = subname $level + 2;
    return defined $sub ? "$sub ($fileline)" : "$fileline";
}

sub uncontrol {
    require 'dumpvar.pl';
    my $r = \&dumpvar::unctrl;
    {
	local $^W = 0;
	*uncontrol = $r;
    }
    goto &$r;
}

1

__END__

=back

=head1 AUTHOR

Roderick Schertler <F<roderick@argon.org>>

=head1 SEE ALSO

infobot(1), perl(1).

=cut
